# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding=utf-8

import os
import sys

parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, parent_dir)

from .http_service import protocol_type as PT
from .http_service import method_type as MT
from .http_service import format_type as FT
from .auth import rpc_signature_composer as rpc_signer
from .auth import roa_signature_composer as roa_signer
from .auth import oss_signature_composer as oss_signer
from .auth import md5_tool
import abc
import base64

"""
Acs request model.

Created on 9/19/2017

@author: guangyao
"""

STYLE_RPC = 'RPC'
STYLE_ROA = 'ROA'
STYLE_OSS = 'OSS'


class AcsRequest:
    """
    Acs request base class. This class wraps up common parameters for a request.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, product, version=None,
                 action_name=None,
                 location_service_code=None,
                 accept_format=None,
                 protocol_type=PT.HTTP,
                 method=None):
        """

        :param product:
        :param version:
        :param action_name:
        :param params:
        :param resource_owner_account:
        :param protocol_type:
        :param accept_format:
        :return:
        """
        self.__version = version
        self.__product = product
        self.__action_name = action_name
        self.__protocol_type = protocol_type
        self.__accept_format = accept_format
        self.__params = {}
        self.__method = method
        self.__header = {}
        self.__uri_pattern = None
        self.__uri_params = None
        self.__content = None
        self.__location_service_code = location_service_code

    def add_query_param(self, k, v):
        if self.__params is None:
            self.__params = {}
        self.__params[k] = v

    def get_uri_pattern(self):
        return self.__uri_pattern

    def get_uri_params(self):
        return self.__uri_params

    def get_product(self):
        return self.__product

    def get_version(self):
        return self.__version

    def get_action_name(self):
        return self.__action_name

    def get_accept_format(self):
        return self.__accept_format

    def get_protocol_type(self):
        return self.__protocol_type

    def get_query_params(self):
        return self.__params

    def get_method(self):
        return self.__method

    def set_uri_pattern(self, pattern):
        self.__uri_pattern = pattern

    def set_uri_params(self, params):
        self.__uri_params = params

    def set_method(self, method):
        self.__method = method

    def set_product(self, product):
        self.__product = product

    def set_version(self, version):
        self.__version = version

    def set_action_name(self, action_name):
        self.__action_name = action_name

    def set_accept_format(self, accept_format):
        self.__accept_format = accept_format

    def set_protocol_type(self, protocol_type):
        self.__protocol_type = protocol_type

    def set_query_params(self, params):
        self.__params = params

    def set_content(self, content):
        """

        :param content: ByteArray
        :return:
        """
        self.__content = content

    def get_content(self):
        """

        :return: ByteArray
        """
        return self.__content

    def get_headers(self):
        """

        :return: Dict
        """
        return self.__header

    def set_headers(self, headers):
        """

        :param headers: Dict
        :return:
        """
        self.__header = headers

    def add_header(self, k, v):
        if self.__header is None:
            self.__header = dict(k=v)
        else:
            self.__header[k] = v

    def set_user_agent(self, agent):
        self.add_header('User-Agent', agent)

    def set_location_service_code(self, location_service_code):
        self.__location_service_code = location_service_code

    def get_location_service_code(self):
        return self.__location_service_code

    @abc.abstractmethod
    def get_style(self):
        pass

    @abc.abstractmethod
    def get_url(self, region_id, ak, secret):
        pass

    @abc.abstractmethod
    def get_signed_header(self, region_id, ak, secret):
        pass


class RpcRequest(AcsRequest):
    """
    Class to compose an RPC style request with.
    """

    def __init__(self, product, version, action_name, location_service_code=None, format=None, protocol=None):
        AcsRequest.__init__(self, product, version, action_name, location_service_code, format, protocol, MT.GET)
        self.__style = STYLE_RPC

    def get_style(self):
        return self.__style

    def __get_sign_params(self):
        req_params = self.get_query_params()
        if req_params is None:
            req_params = {}
        req_params['Version'] = self.get_version()
        req_params['Action'] = self.get_action_name()
        req_params['Format'] = self.get_accept_format()
        return req_params

    def get_url(self, region_id, ak, secret):
        sign_params = self.__get_sign_params()
        if 'RegionId' not in sign_params.keys():
            sign_params['RegionId'] = region_id
        url = rpc_signer.get_signed_url(sign_params, ak, secret, self.get_accept_format(), self.get_method())
        return url

    def get_signed_header(self, region_id=None, ak=None, secret=None):
        return {}


class RoaRequest(AcsRequest):
    """
    Class to compose an ROA style request with.
    """

    def __init__(self, product, version, action_name, location_service_code=None, method=None, headers=None, uri_pattern=None, path_params=None,
                 protocol=None):
        """

        :param product: String, mandatory
        :param version: String, mandatory
        :param action_name: String, mandatory
        :param method: String
        :param headers: Dict
        :param uri_pattern: String
        :param path_params: Dict
        :param protocol: String
        :return:
        """
        AcsRequest.__init__(self, product, version, action_name, location_service_code, FT.RAW, protocol, method)
        self.__style = STYLE_ROA
        self.__method = method
        self.__header = headers
        self.__uri_pattern = uri_pattern
        self.__path_params = path_params

    def get_style(self):
        """

        :return: String
        """
        return self.__style

    def get_path_params(self):
        return self.__path_params

    def set_path_params(self, path_params):
        self.__path_params = path_params

    def add_path_param(self, k, v):
        if self.__path_params is None:
            self.__path_params = {}
        self.__path_params[k] = v

    def __get_sign_params(self):
        req_params = self.get_query_params()
        if req_params is None:
            req_params = {}
        req_params['Version'] = self.get_version()
        req_params['Action'] = self.get_action_name()
        req_params['Format'] = self.get_accept_format()
        return req_params

    def get_signed_header(self, region_id, ak, secret):
        """
        Generate signed header
        :param region_id: String
        :param ak: String
        :param secret: String
        :return: Dict
        """
        sign_params = self.get_query_params()
        if (self.get_content() is not None):
            md5_str = md5_tool.get_md5_base64_str(self.get_content())
            self.add_header('Content-MD5', md5_str)
        if 'RegionId' not in sign_params.keys():
            sign_params['RegionId'] = region_id
        signed_headers = roa_signer.get_signature_headers(sign_params, ak, secret,
                                                          self.get_accept_format(),
                                                          self.get_headers(),
                                                          self.get_uri_pattern(),
                                                          self.get_path_params(),
                                                          self.get_method())
        return signed_headers

    def get_url(self, region_id, ak=None, secret=None):
        """
        Compose request url without domain
        :param region_id: String
        :return: String
        """
        sign_params = self.get_query_params()
        if region_id not in sign_params.keys():
            sign_params['RegionId'] = region_id
        url = roa_signer.get_url(self.get_uri_pattern(), sign_params, self.get_path_params())
        return url


class OssRequest(AcsRequest):
    def __init__(self, product, version, action_name, location_service_code, bucket=None, method=None,
                 headers=None, uri_pattern=None, path_params=None, protocol=None):
        """

        :param product: String, mandatory
        :param version: String, mandatory
        :param action_name: String, mandatory
        :param bucket: String
        :param method: String
        :param headers: Dict
        :param uri_pattern: String
        :param path_params: Dict
        :param protocol: String
        :return:
        """
        AcsRequest.__init__(self, product, version, action_name, location_service_code, FT.XML, protocol, method)
        self.__style = STYLE_OSS
        self.__bucket = bucket
        self.__method = method
        self.__header = headers
        self.__uri_pattern = uri_pattern
        self.__path_params = path_params

    def get_style(self):
        return self.__style

    def get_path_params(self):
        """

        :return: dict
        """
        return self.__path_params

    def set_path_params(self, path_params):
        self.__path_params = path_params

    def add_path_param(self, k, v):
        if self.__path_params is None:
            self.__path_params = {}
        self.__path_params[k] = v

    def __get_sign_params(self):
        req_params = self.get_query_params()
        if req_params is None:
            req_params = {}
        req_params['Version'] = self.get_version()
        req_params['Action'] = self.get_action_name()
        req_params['Format'] = self.get_accept_format()
        return req_params

    def get_signed_header(self, region_id, ak, secret, ):
        """
        Compose signed headers.
        :param region_id: String
        :param ak: String
        :param secret: String
        :return:
        """
        sign_params = self.get_query_params()
        if 'RegionId' not in sign_params.keys():
            sign_params['RegionId'] = region_id
        signed_headers = oss_signer.get_signature_headers(sign_params, ak, secret, self.get_accept_format(),
                                                          self.get_headers(),
                                                          self.get_uri_pattern(), self.get_path_params(),
                                                          self.get_method(), self.__bucket)
        return signed_headers

    def get_url(self, region_id, ak=None, secret=None):
        """
        Generate request url without domain
        :param region_id: String
        :return: String
        """
        sign_params = self.get_query_params()
        if 'RegionId' not in sign_params.keys():
            sign_params['RegionId'] = region_id
        url = oss_signer.get_url(sign_params, self.get_uri_pattern(), self.get_path_params())
        return url



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#coding=utf-8

import os
import sys

parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, parent_dir)
from ..acs_exception import error_code, error_msg
from ..acs_exception.exceptions import ClientException
from xml.dom.minidom import parse


"""
Region&Endpoint provider module.

Created on 9/19/2017

@author: guangyao
"""

#endpoint list
__endpoints = dict()

#load endpoints info from endpoints.xml file and parse to dict.
__endpoints_file = os.path.join(parent_dir, 'endpoints.xml')
try:
    DOMTree = parse(__endpoints_file)
    root = DOMTree.documentElement
    eps = root.getElementsByTagName('Endpoint')
    for endpoint in eps:
        region_list = []
        product_list = []
        regions = endpoint.getElementsByTagName('RegionId')
        products = endpoint.getElementsByTagName('Product')
        for region in regions:
            region_list.append(region.childNodes[0].nodeValue)
        for product in products:
            name_node = product.getElementsByTagName('ProductName')[0]
            name = name_node.childNodes[0].nodeValue
            domain_node = product.getElementsByTagName('DomainName')[0]
            domain = domain_node.childNodes[0].nodeValue
            product_list.append({name: domain})

        __endpoints[endpoint.getAttribute('name')] = dict(regions=region_list, products=product_list)

except Exception as ex:
    raise ClientException(error_code.SDK_MISSING_ENDPOINTS_FILER, error_msg.get_msg('SDK_MISSING_ENDPOINTS_FILER'))


def find_product_domain(regionid, prod_name):
    """
	Fetch endpoint url with given region id, product name and endpoint list
	:param regionid: region id
	:param product: product name
	:param endpoints: product list
	:return: endpoint url
	"""
    if regionid is not None and product is not None:
        for point in __endpoints:
            point_info = __endpoints.get(point)
            if regionid in point_info.get('regions'):
                prod_info = point_info.get('products')
                for prod in prod_info:
                    if prod_name in prod:
                        return prod.get(prod_name)
    return None

def modify_point(product_name, region_id, end_point):
    for point in __endpoints:
        point_info = __endpoints.get(point)
        region_list = point_info.get('regions')
        products = point_info.get('products')

        if region_id is not None and region_id not in region_list:
            region_list.append(region_id)

        if end_point is not None:
            product_exit = 0
            for prod in products:
                if product_name in prod:
                    prod[product_name] = end_point
                    product_exit = 1
            if product_exit == 0:
                item = dict()
                item[product_name] = end_point
                products.append(item)

        __mdict = dict()
        __mdict['regions'] = region_list
        __mdict['products'] = products
        __endpoints[point] = __mdict
        convert_dict_to_endpointsxml(__endpoints)

def convert_dict_to_endpointsxml(mdict):
    regions = list()
    products = list()
    for point in mdict:
        point_info = mdict.get(point)
        regions = point_info.get('regions')
        products = point_info.get('products')
    content = ''
    prefix = '<?xml version="1.0" encoding="UTF-8"?>\n<Endpoints>\n<Endpoint name="cn-hangzhou">\n'
    endfix = '</Endpoint>\n</Endpoints>\n'
    content += prefix
    content += '<RegionIds>\n'
    for item in regions:
        content += '<RegionId>'+item+'</RegionId>\n'
    content += '</RegionIds>\n'+'<Products>\n'
    for item in products:
        content += '<Product>\n'
        content += '<ProductName>'+list(item.keys())[0]+'</ProductName>\n'
        content += '<DomainName>'+item[list(item.keys())[0]]+'</DomainName>\n'
        content += '</Product>\n'
    content += '</Products>'
    content += endfix
    #print content
    if not os.path.isfile(__endpoints_file):
        _createFile(__endpoints_file)
    f = open(__endpoints_file, 'w')
    try:
        f.write(''.join(content))
    except Exception as e:
        print(e)
        print("Please confirm you has use sudo + cmd")
    finally:
        f.close()

def _createFile(filename):
    namePath = os.path.split(filename)[0]
    if not os.path.isdir(namePath):
        os.makedirs(namePath)
        with os.fdopen(os.open(filename,
                               os.O_WRONLY | os.O_CREAT, 0o600), 'w'):
            pass

if __name__ == '__main__':
    print(find_product_domain('cn-hangzhou', 'Rds'))
    modify_point('ecs', 'cn-beijing-2', 'ecs.aliyuncs.com')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding=utf-8
import os
import sys
import http.client as httplib
import warnings
warnings.filterwarnings("once", category=DeprecationWarning)

try:
    import json
except ImportError:
    import simplejson as json

from .profile import region_provider
from .profile.location_service import LocationService
from .acs_exception.exceptions import ClientException
from .acs_exception.exceptions import ServerException
from .acs_exception import error_code, error_msg
from .http_service.http_response import HttpResponse
from .request import AcsRequest

"""
Acs default client module.

Created on 9/19/2017

@author: guangyao
"""


class AcsClient:
    def __init__(self, ak, secret, region_id, auto_retry=True, max_retry_time=3, user_agent=None, port=80):
        """
        constructor for AcsClient
        :param ak: String, access key id
        :param secret: String, access key secret
        :param region_id: String, region id
        :param auto_retry: Boolean
        :param max_retry_time: Number
        :return:
        """
        self.__max_retry_num = max_retry_time
        self.__auto_retry = auto_retry
        self.__ak = ak
        self.__secret = secret
        self.__region_id = region_id
        self.__user_agent = user_agent
        self._port = port
        self._location_service = LocationService(self)
        self._url_test_flag = False # if true, do_action() will throw a ClientException that contains URL

    def get_region_id(self):
        """

        :return: String
        """
        return self.__region_id

    def get_access_key(self):
        """

        :return: String
        """
        return self.__ak

    def get_access_secret(self):
        """

        :return: String
        """
        return self.__secret

    def is_auto_retry(self):
        """

        :return:Boolean
        """
        return self.__auto_retry

    def get_max_retry_num(self):
        """

        :return: Number
        """
        return self.__max_retry_num

    def get_user_agent(self):
        return self.__user_agent

    def set_region_id(self, region):
        self.__region_id = region

    def set_access_key(self, ak):
        self.__ak = ak

    def set_access_secret(self, secret):
        self.__secret = secret

    def set_max_retry_num(self, num):
        """
        set auto retry number
        :param num: Numbers
        :return: None
        """
        self.__max_retry_num = num

    def set_auto_retry(self, flag):
        """
        set whether or not the client perform auto-retry
        :param flag: Booleans
        :return: None
        """
        self.__auto_retry = flag

    def set_user_agent(self, agent):
        """
        User agent set to client will overwrite the request setting.
        :param agent:
        :return:
        """
        self.__user_agent = agent

    def get_location_service(self):
        return self._location_service

    def get_port(self):
        return self._port

    def _resolve_endpoint(self, request):
        endpoint = None
        if request.get_location_service_code() is not None:
            endpoint = self._location_service.find_product_domain(self.get_region_id(), request.get_location_service_code())
        if endpoint is None:
            endpoint = region_provider.find_product_domain(self.get_region_id(), request.get_product())
            if endpoint is None:
                raise ClientException(error_code.SDK_INVALID_REGION_ID, error_msg.get_msg('SDK_INVALID_REGION_ID'))
            if not isinstance(request, AcsRequest):
                raise ClientException(error_code.SDK_INVALID_REQUEST, error_msg.get_msg('SDK_INVALID_REQUEST'))
        return endpoint

    def _make_http_response(self, endpoint, request):
        content = request.get_content()
        method = request.get_method()
        header = request.get_signed_header(self.get_region_id(), self.get_access_key(),
                                               self.get_access_secret())
        if self.get_user_agent() is not None:
            header['User-Agent'] = self.get_user_agent()
            header['x-sdk-client'] = 'python/2.0.0'
        if header is None:
            header = {}

        protocol = request.get_protocol_type()
        url = request.get_url(self.get_region_id(), self.get_access_key(), self.get_access_secret())
        response = HttpResponse(endpoint, url, method, header, protocol, content,
                                self._port)
        return response

    def _implementation_of_do_action(self, request):
        endpoint = self._resolve_endpoint(request)
        http_response = self._make_http_response(endpoint, request)
        if self._url_test_flag:
            raise ClientException("URLTestFlagIsSet", http_response.get_url())

        # Do the actual network thing
        try:
            status, headers, body = http_response.get_response_object()
            return status, headers, body
        except IOError as e:
            raise ClientException(error_code.SDK_SERVER_UNREACHABLE, error_msg.get_msg('SDK_SERVER_UNREACHABLE') + ': ' + str(e))
        except AttributeError:
            raise ClientException(error_code.SDK_INVALID_REQUEST, error_msg.get_msg('SDK_INVALID_REQUEST'))

    def _parse_error_info_from_response_body(self, response_body):
        try:
            body_obj = json.loads(response_body)
            if 'Code' in body_obj and 'Message' in body_obj:
                return (body_obj['Code'], body_obj['Message'])
            else:
                return (error_code.SDK_UNKNOWN_SERVER_ERROR, error_msg.get_msg('SDK_UNKNOWN_SERVER_ERROR'))
        except ValueError:
            # failed to parse body as json format
            return (error_code.SDK_UNKNOWN_SERVER_ERROR, error_msg.get_msg('SDK_UNKNOWN_SERVER_ERROR'))

    def do_action_with_exception(self, acs_request):

        # set server response format as json, because thie function will
        # parse the response so which format doesn't matter
        acs_request.set_accept_format('json')

        status, headers, body = self._implementation_of_do_action(acs_request)

        request_id = None
        ret = body

        try:
            body_obj = json.loads(str(body))
            request_id = body_obj.get('RequestId')
            ret = body_obj
        except ValueError:
            # in case the response body is not a json string, return the raw data instead
            pass

        if status != 200:
            server_error_code, server_error_message = self._parse_error_info_from_response_body(body)
            raise ServerException(server_error_code, server_error_message, http_status=status, request_id=request_id)

        return body

    def do_action(self, acs_request):
        warnings.warn("do_action() method is deprecated, please use do_action_with_exception() instead.", DeprecationWarning)
        status, headers, body = self._implementation_of_do_action(acs_request)
        return body

    def get_response(self, acs_request):
        warnings.warn("get_response() method is deprecated, please use do_action_with_exception() instead.", DeprecationWarning)
        return self._implementation_of_do_action(acs_request)




# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#coding=utf-8
from urllib import parse
import sys

"""
Acs url encoder module.

Created on 9/19/2017

@author: guangyao
"""


def get_encode_str(params):
	"""
	transforms parameters to encoded string
	:param params: dict parameters
	:return: string
	"""
	list_params = sorted(params.iteritems(), key=lambda d: d[0])
	encode_str = parse.urlencode(list_params)
	if sys.stdin.encoding is None:
		res = parse.quote(encode_str.decode('cp936').encode('utf8'), '')
	else:
		res = parse.quote(encode_str.decode(sys.stdin.encoding).encode('utf8'), '')
	res = res.replace("+","%20")
	res = res.replace("*","%2A")
	res = res.replace("%7E","~")
	return res


# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#coding=utf-8
__author__='guangyao'
import http.client as httplib

from .http_request import HttpRequest
from . import protocol_type as PT


class HttpResponse(HttpRequest):
	def __init__(self,host="",url="/",method="GET",headers={},protocol=PT.HTTP,content= None, port=None,key_file=None,cert_file=None):
		HttpRequest.__init__(self,host=host,url=url,method=method,headers=headers)
		self.__ssl_enable=False
		if protocol is PT.HTTPS:
			self.__ssl_enable=True
		self.__key_file=key_file
		self.__cert_file=cert_file
		self.__port=port
		self.__connection=None
		self.set_body(content)

	def set_ssl_enable(self,enable):
		self.__ssl_enable=enable

	def get_ssl_enabled(self):
		return self.__ssl_enable

	def get_response(self):
		if self.get_ssl_enabled():
			return self.get_https_response()
		else:
			return self.get_http_response()


	def get_response_object(self):
		if self.get_ssl_enabled():
			return self.get_https_response_object()
		else:
			return self.get_http_response_object()


	def get_http_response(self):
		if self.__port is None or self.__port == "":
			self.__port = 80
		try:
			self.__connection=httplib.HTTPConnection(self.get_host(),self.__port)
			self.__connection.connect()
			self.__connection.request(method=self.get_method(),url=self.get_url(),body=self.get_body(),
			                          headers=self.get_headers())
			response=self.__connection.getresponse()
			return response.getheaders(), response.read()
		finally:
			self.__close_connection()


	def get_http_response_object(self):
		if self.__port is None or self.__port == "":
			self.__port = 80
		try:
			self.__connection=httplib.HTTPConnection(self.get_host(),self.__port)
			self.__connection.connect()
			self.__connection.request(method=self.get_method(),url=self.get_url(),body=self.get_body(),
			                          headers=self.get_headers())
			response=self.__connection.getresponse()
			return response.status, response.getheaders(), response.read()
		finally:
			self.__close_connection()


	def get_https_response(self):
		if self.__port is None or self.__port == "":
			self.__port = 443
		try:
			self.__port = 443
			self.__connection=httplib.HTTPSConnection(self.get_host(),self.__port,cert_file=self.__cert_file,
			                                          key_file=self.__key_file)
			self.__connection.connect()
			self.__connection.request(method=self.get_method(),url=self.get_url(),body=self.get_body(),
			                          headers=self.get_headers())
			response=self.__connection.getresponse()
			return response.getheaders(), response.read()
		finally:
			self.__close_connection()


	def get_https_response_object(self):
		if self.__port is None or self.__port == "":
			self.__port = 443
		try:
			self.__port = 443
			self.__connection=httplib.HTTPSConnection(self.get_host(),self.__port,cert_file=self.__cert_file,
			                                          key_file=self.__key_file)
			self.__connection.connect()
			self.__connection.request(method=self.get_method(),url=self.get_url(),body=self.get_body(),
			                          headers=self.get_headers())
			response=self.__connection.getresponse()
			return response.status, response.getheaders(), response.read()
		finally:
			self.__close_connection()


	def __close_connection(self):
		if self.__connection is not None:
			self.__connection.close()
			self.__connection = None





# Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
 # distributed with this work for additional information
 # regarding copyright ownership.  The ASF licenses this file
 # to you under the Apache License, Version 2.0 (the
 # "License"); you may not use this file except in compliance
 # with the License.  You may obtain a copy of the License at
 #
 #     http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing,
 # software distributed under the License is distributed on an
 # "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 # KIND, either express or implied.  See the License for the
 # specific language governing permissions and limitations
 # under the License.

 #coding=utf-8

__author__ = 'guangyao'
import os
import sys

parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, parentdir)
from . import sha_hmac1 as mac1
from ..utils import parameter_helper as helper
from ..http_service import format_type as FormatType
from urllib import parse

ACCEPT = "Accept"
CONTENT_MD5 = "Content-MD5"
CONTENT_TYPE = "Content-Type"
DATE = "Date"
QUERY_SEPARATOR = "&"
HEADER_SEPARATOR = "\n"

def __init__():
    pass

# this function will append the necessary parameters for signer process.
# parameters: the orignal parameters
# signer: sha_hmac1 or sha_hmac256
# accessKeyId: this is aliyun_access_key_id
# format: XML or JSON
# input parameters is headers
def refresh_sign_parameters(parameters, access_key_id, format, signer=mac1):
    if parameters is None or not isinstance(parameters, dict):
        parameters = dict()
    if format is None:
        format = FormatType.RAW
    parameters["Date"] = helper.get_rfc_2616_date()
    parameters["Accept"] = FormatType.map_format_to_accept(format)
    parameters["x-acs-signature-method"] = signer.get_signer_name()
    parameters["x-acs-signature-version"] = signer.get_singer_version()
    return parameters


def compose_string_to_sign(method, queries, uri_pattern=None, headers=None, paths=None, signer=mac1):
    sign_to_string = ""
    sign_to_string += method
    sign_to_string += HEADER_SEPARATOR
    if headers.has_key(ACCEPT) and headers[ACCEPT] is not None:
        sign_to_string += headers[ACCEPT]
    sign_to_string += HEADER_SEPARATOR
    if headers.has_key(CONTENT_MD5) and headers[CONTENT_MD5] is not None:
        sign_to_string += headers[CONTENT_MD5]
    sign_to_string += HEADER_SEPARATOR
    if headers.has_key(CONTENT_TYPE) and headers[CONTENT_TYPE] is not None:
        sign_to_string += headers[CONTENT_TYPE]
    sign_to_string += HEADER_SEPARATOR
    if headers.has_key(DATE) and headers[DATE] is not None:
        sign_to_string += headers[DATE]
    sign_to_string += HEADER_SEPARATOR
    uri = replace_occupied_parameters(uri_pattern, paths)
    sign_to_string += build_canonical_headers(headers, "x-acs-")
    sign_to_string += __build_query_string(uri, queries)
    return sign_to_string

def replace_occupied_parameters(uri_pattern, paths):
    result = uri_pattern
    if paths is not None:
        for (key, value) in paths.items():
            target = "[" + key + "]"
            result = result.replace(target, value)
    return result

# change the give headerBegin to the lower() which in the headers
# and change it to key.lower():value
def build_canonical_headers(headers, header_begin):
    result = ""
    unsort_map = dict()
    for (key, value) in headers.iteritems():
        if key.lower().find(header_begin) >= 0:
            unsort_map[key.lower()] = value
    sort_map = sorted(unsort_map.items(), key=lambda d: d[0])
    for (key, value) in sort_map:
        result += key + ":" +value
        result += HEADER_SEPARATOR
    return result

def split_sub_resource(uri):
    return uri.split("?")

def __build_query_string(uri, queries):
    uri_parts = split_sub_resource(uri)
    if len(uri_parts) >1 and uri_parts[1] is not None:
        queries[uri_parts[1]] = None
    query_builder = uri_parts[0]
    sorted_map = sorted(queries.items(), key=lambda queries:queries[0])
    if len(sorted_map) > 0:
        query_builder += "?"
    for (k,v) in sorted_map:
        query_builder += k
        if v is not None:
            query_builder += "="
            query_builder += str(v)
        query_builder += QUERY_SEPARATOR
    if query_builder.endswith(QUERY_SEPARATOR):
        query_builder = query_builder[0:(len(query_builder)-1)]
    return query_builder

def get_signature(queries, access_key, secret, format, headers, uri_pattern, paths, method, signer=mac1):
    headers = refresh_sign_parameters(parameters=headers, access_key_id=access_key,format=format)
    sign_to_string = compose_string_to_sign(method=method, queries=queries, headers=headers, uri_pattern=uri_pattern, paths=paths)
    signature = signer.get_sign_string(sign_to_string, secret=secret)
    return signature

def get_signature_headers(queries, access_key, secret, format, headers, uri_pattern, paths, method, signer=mac1):
    signature = get_signature(queries, access_key, secret, format, headers, uri_pattern, paths, method, signer)
    headers["Authorization"] = "acs "+access_key+":"+signature
    return headers

def get_url(uri_pattern, queries, path_parameters):
    url = ""
    url += replace_occupied_parameters(uri_pattern, path_parameters)
    if not url.endswith("?"):
        url += "?"
    url += parse.urlencode(queries)
    if url.endswith("?"):
        url = url[0:(len(url)-1)]
    return url



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#coding=utf-8

"""
Acs error message module.

Created on 9/19/2017

@author: guangyao
"""

__dict = dict(SDK_INVALID_REGION_ID='Can not find endpoint to access.',
              SDK_SERVER_UNREACHABLE='Unable to connect server',
              SDK_INVALID_REQUEST='The request is not a valid AcsRequest.',
              SDK_MISSING_ENDPOINTS_FILER='Internal endpoints info is missing.',
              SDK_UNKNOWN_SERVER_ERROR="Can not parse error message from server response.")


def get_msg(code):
	return __dict.get(code)



 # Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
 # distributed with this work for additional information
 # regarding copyright ownership.  The ASF licenses this file
 # to you under the Apache License, Version 2.0 (the
 # "License"); you may not use this file except in compliance
 # with the License.  You may obtain a copy of the License at
 #
 #     http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing,
 # software distributed under the License is distributed on an
 # "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 # KIND, either express or implied.  See the License for the
 # specific language governing permissions and limitations
 # under the License.

 #coding=utf-8

__author__ = 'guangyao'

XML  = 'XML'
JSON = 'JSON'
RAW  = 'RAW'
APPLICATION_XML = 'application/xml'
APPLICATION_JSON = 'application/json'
APPLICATION_OCTET_STREAM = 'application/octet-stream'
TEXT_XML = 'text/xml'

def map_format_to_accept(format):
    if format == XML:
        return APPLICATION_XML
    if format == JSON:
        return APPLICATION_JSON
    return APPLICATION_OCTET_STREAM

def map_accept_to_format(accept):
    if accept.lower() == APPLICATION_XML or accept.lower() == TEXT_XML:
        return XML
    if accept.lower() == APPLICATION_JSON:
        return JSON
    return RAW

if __name__ == "__main__":
    print(map_format_to_accept(XML))
    print(map_format_to_accept(JSON))
    print(map_format_to_accept(RAW))
    print(map_accept_to_format("application/xml"))
    print(map_accept_to_format("text/xml"))
    print(map_accept_to_format("application/json"))





# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#coding=utf-8

import os
import sys
import json
parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, parent_dir)

from ..request import RpcRequest
from ..http_service.http_response import HttpResponse
from ..acs_exception import exceptions as exs
from ..acs_exception import error_code, error_msg

LOCATION_SERVICE_PRODUCT_NAME="Location"
LOCATION_SERVICE_DOMAIN="location.aliyuncs.com"
LOCATION_SERVICE_VERSION="2015-06-12"
LOCATION_SERVICE_DESCRIBE_ENDPOINT_ACTION="DescribeEndpoint"
LOCATION_SERVICE_REGION="cn-hangzhou"

class DescribeEndpointRequest(RpcRequest):

    def __init__(self, product_name, version, action_name, region_id, service_code):
        RpcRequest.__init__(self, product_name, version, action_name, 'hhh')

        self.add_query_param("Id", region_id)
        self.add_query_param("ServiceCode", service_code)
        self.set_accept_format("JSON")


class LocationService:

    def __init__(self, client):
        self.__clinetRef = client
        self.__cache = {}
        self.__service_product_name = LOCATION_SERVICE_PRODUCT_NAME
        self.__service_domain = LOCATION_SERVICE_DOMAIN
        self.__service_version = LOCATION_SERVICE_VERSION
        self.__service_region = LOCATION_SERVICE_REGION
        self.__service_action = LOCATION_SERVICE_DESCRIBE_ENDPOINT_ACTION

    def set_location_service_attr(self, region=None, product_name=None, domain=None,version=None):
        if region is not None:
            self.__service_region = region

        if product_name is not None:
            self.__service_product_name = product_name

        if domain is not None:
            self.__service_domain = domain

        if version is not None:
            self.__service_version = version

    def find_product_domain(self, region_id, service_code):
        key = "%s_&_%s" %(region_id, service_code)
        domain = self.__cache.get(key)
        if domain is None:
            domain = self.find_product_domain_from_location_service(region_id, service_code)
            if domain is not None:
                self.__cache[key] = domain

        return domain


    def find_product_domain_from_location_service(self, region_id, service_code):

        request = DescribeEndpointRequest(self.__service_product_name,
                                           self.__service_version,
                                           self.__service_action,
                                           region_id,
                                           service_code)
        try:
            content = request.get_content()
            method = request.get_method()
            header = request.get_signed_header(self.__service_region, self.__clinetRef.get_access_key(),
                                                   self.__clinetRef.get_access_secret())
            if self.__clinetRef.get_user_agent() is not None:
                header['User-Agent'] = self.__clinetRef.get_user_agent()
                header['x-sdk-client'] = 'python/2.0.0'
            protocol = request.get_protocol_type()
            url = request.get_url(self.__service_region, self.__clinetRef.get_access_key(),
                                  self.__clinetRef.get_access_secret())
            response = HttpResponse(self.__service_domain, url, method, {} if header is None else header, protocol, content,
                                    self.__clinetRef.get_port())

            status, header, body = response.get_response_object()
            result = json.loads(body)
            if status == 200:
                return result.get('Endpoint')
            elif status >= 400 and status < 500:
                # print "serviceCode=" + service_code + " get location error! code=" + result.get('Code') +", message =" + result.get('Message')
                return None
            elif status >= 500:
                raise exs.ServerException(result.get('Code'), result.get('Message'))
            else:
                raise exs.ClientException(result.get('Code'), result.get('Message'))
        except IOError:
            raise exs.ClientException(error_code.SDK_SERVER_UNREACHABLE, error_msg.get_msg('SDK_SERVER_UNREACHABLE'))
        except AttributeError:
            raise exs.ClientException(error_code.SDK_INVALID_REQUEST, error_msg.get_msg('SDK_INVALID_REQUEST'))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

#coding=utf-8

"""
MD5 tools module.

Created on 9/19/2017

@author: guangyao
"""

import hashlib
import base64



def _get_md5(content):
    m = hashlib.md5()
    m.update(memoryview(content))
    return m.digest()

def get_md5_base64_str(content):
    return base64.encodestring(_get_md5(content)).strip()



# Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
 # distributed with this work for additional information
 # regarding copyright ownership.  The ASF licenses this file
 # to you under the Apache License, Version 2.0 (the
 # "License"); you may not use this file except in compliance
 # with the License.  You may obtain a copy of the License at
 #
 #     http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing,
 # software distributed under the License is distributed on an
 # "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 # KIND, either express or implied.  See the License for the
 # specific language governing permissions and limitations
 # under the License.

 #coding=utf-8

__author__ = 'guangyao'

import hmac
import hashlib
import base64

class ShaHmac256:
    def __init__(self):
        pass

    def get_sign_string(self, source, accessSecret):
        h = hmac.new(accessSecret, source, hashlib.sha256)
        signature = base64.encodestring(h.digest()).strip()
        return signature

    def get_signer_name(self):
        return "HMAC-SHA256"

    def get_singer_version(self):
        return "1.0"



 # Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
 # distributed with this work for additional information
 # regarding copyright ownership.  The ASF licenses this file
 # to you under the Apache License, Version 2.0 (the
 # "License"); you may not use this file except in compliance
 # with the License.  You may obtain a copy of the License at
 #
 #     http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing,
 # software distributed under the License is distributed on an
 # "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 # KIND, either express or implied.  See the License for the
 # specific language governing permissions and limitations
 # under the License.

 #coding=utf-8

__author__ = 'guangyao'

GET = "GET"
PUT = "PUT"
POST = "POST"
DELETE = "DELETE"
HEAD = "HEAD"
OPTIONS = "OPTIONS"



 # Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
 # distributed with this work for additional information
 # regarding copyright ownership.  The ASF licenses this file
 # to you under the Apache License, Version 2.0 (the
 # "License"); you may not use this file except in compliance
 # with the License.  You may obtain a copy of the License at
 #
 #     http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing,
 # software distributed under the License is distributed on an
 # "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 # KIND, either express or implied.  See the License for the
 # specific language governing permissions and limitations
 # under the License.

 #coding=utf-8

__author__ = 'guangyao'
import os
import sys

parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0,parentdir)
from . import format_type
from ..utils import parameter_helper as helper


class HttpRequest:

    content_md5 = "Content-MD5"
    content_length = "Content-Length"
    content_type = "Content-Type"

    def __init__(self, host="", url="/", method=None, headers={}):
        self.__host = host
        self.__url = url
        self.__method = method
        self.__content_type = ""
        self.__content = ""
        self.__encoding = ""
        self.__headers = headers
        self.__body = None

    def get_host(self):
        return self.__host

    def set_host(self, host):
        self.__host = host

    def get_body(self):
        return self.__body

    def set_body(self, body):
        self.__body = body

    def get_url(self):
        return self.__url

    def set_url(self, url):
        self.__url = url

    def get_encoding(self):
        return self.__encoding

    def set_encoding(self, encoding):
        self.__encoding = encoding

    def get_content_type(self):
        return self.__content_type

    def set_content_type(self, content_type):
        self.__content_type = content_type

    def get_method(self):
        return self.__method

    def set_method(self, method):
        self.__method = method

    def get_content(self):
        return self.__content

    def get_header_value(self, name):
        return self.__headers[name]

    def put_header_parameter(self, key, value):
        if key is not None and value is not None:
            self.__headers[key] = value

    def md5_sum(self, content):
        return helper.md5_sum(content)

    def set_content(self, content, encoding, format):
        tmp = dict()
        if content is None:
            self.__headers.pop(self.content_md5)
            self.__headers.pop(self.content_length)
            self.__headers.pop(self.content_type)
            self.__content_type = None
            self.__content = None
            self.__encoding = None
            return
        str_md5 = self.md5_sum(content)
        content_length = len(content)
        content_type = format_type.RAW
        if format is None:
            content_type = format
        self.__headers[self.content_md5] = str_md5
        self.__headers[self.content_length] = content_length
        self.__headers[self.content_type] = content_type
        self.__content = content
        self.__encoding = encoding

    def get_headers(self):
        return self.__headers






# Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
 # distributed with this work for additional information
 # regarding copyright ownership.  The ASF licenses this file
 # to you under the Apache License, Version 2.0 (the
 # "License"); you may not use this file except in compliance
 # with the License.  You may obtain a copy of the License at
 #
 #     http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing,
 # software distributed under the License is distributed on an
 # "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 # KIND, either express or implied.  See the License for the
 # specific language governing permissions and limitations
 # under the License.

 #coding=utf-8

__author__ = 'guangyao'
import os
import sys

parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0,parentdir)
from . import roa_signature_composer
from . import sha_hmac1 as mac1
from ..utils import parameter_helper as helper
from urllib import parse

ACCEPT = "Accept"
CONTENT_MD5 = "Content-MD5"
CONTENT_TYPE = "Content-Type"
DATE = "Date"
QUERY_SEPARATOR = "&"
HEADER_SEPARATOR = "\n"

def __init__():
    pass

def refresh_sign_parameters(parameters, access_key_id, format="JSON", signer=mac1):
    parameters["Date"] = helper.get_rfc_2616_date()
    return parameters

def __build_query_string(uri, queries):
    sorted_map = sorted(queries.items(), key=lambda queries:queries[0])
    if len(sorted_map) > 0:
        uri += "?"
        for (k,v) in sorted_map:
            uri += k
            if v is not None:
                uri += "="
                uri += v
            uri += roa_signature_composer.QUERY_SEPARATOR
    if uri.find(roa_signature_composer.QUERY_SEPARATOR) >= 0:
        uri = uri[0:(len(uri)-1)]
    return uri

def compose_string_to_sign(method, queries, uri_pattern=None, headers=None, paths=None, signer=mac1):
    sign_to_string = ""
    sign_to_string += method
    sign_to_string += HEADER_SEPARATOR
    if headers.has_key(CONTENT_MD5) and headers[CONTENT_MD5] is not None:
        sign_to_string += headers[CONTENT_MD5]
    sign_to_string += HEADER_SEPARATOR
    if headers.has_key(CONTENT_TYPE) and headers[CONTENT_TYPE] is not None:
        sign_to_string += headers[CONTENT_TYPE]
    sign_to_string += HEADER_SEPARATOR
    if headers.has_key(DATE) and headers[DATE] is not None:
        sign_to_string += headers[DATE]
    sign_to_string += HEADER_SEPARATOR
    sign_to_string += roa_signature_composer.build_canonical_headers(headers, "x-oss-")
    sign_to_string += __build_query_string(uri_pattern, queries)
    return sign_to_string

def get_signature(queries, access_key, secret, format, headers, uri_pattern, paths, method, signer=mac1, bucket_name=None):
    headers = refresh_sign_parameters(parameters=headers, access_key_id=access_key,format=format)
    uri = uri_pattern
    if bucket_name is not None:
        uri = "/"+bucket_name+uri
    sign_to_string = compose_string_to_sign(method=method, queries=queries, headers=headers, uri_pattern=uri, paths=paths)
    signature = signer.get_sign_string(sign_to_string, secret=secret)
    return signature

def get_signature_headers(queries, access_key, secret, format, headers, uri_pattern, paths, method, bucket_name, signer=mac1):
    signature = get_signature(queries, access_key, secret, format, headers, uri_pattern, paths, method, signer, bucket_name)
    headers["Authorization"] = "OSS "+access_key+":"+signature
    return headers

def get_url(queries, uri_pattern, path_parameters):
    url = ""
    url += roa_signature_composer.replace_occupied_parameters(uri_pattern, path_parameters)
    if not url.endswith("?"):
        url += "?"
    url += parse.urlencode(queries)
    if url.endswith("?"):
        url = url[0:(len(url)-1)]
    return url